unit birthd_core;
interface
uses SysUtils, Windows;

type TDate=record y,m,d:integer; end;

//{$DEFINE DEBUGDAY}
//const OverrideDay:TDate=(y:2002;m:6;d:15);

type TParserMessageProc=procedure(const msg:string);
var ParserMessageProc:TParserMessageProc;

type TParserNotifyConfigProc=function(const Key,ValueS:string;ValueI:integer;IsInt:boolean):boolean;
var ParserNotifyConfigProc:TParserNotifyConfigProc;

type TBirthday=record
                 Diff:integer;
                 Line:string[124];
               end;
var Birthdays:array of TBirthday;

procedure Parse(const FileName:string);

////////////////////////////////////////////////////////////////////////////////
implementation

var WarnInterval:integer=7;
    PrintSectionName:integer=0;

var LC_LINE_DEFAULT:string='$name turns $age $date';
    LC_LINE_NO_AGE:string='$name''s birthday is $date';
    LC_TODAY:string='today';
    LC_TOMORROW:string='tomorrow';
    LC_THIS_WEEK:string='$weekday';
    LC_WEEKDAYS:array[0..6]of string[15]=
      ('Sunday','Monday','Tuesday','Wednesday','Thursday','Friday','Saturday');
    LC_NEXT_WEEK:string='next week';
    LC_SOMEDAY:string='in $days days';

////////////////////////////////////////////////////////////////////////////////
var Today:TDate;
    TodayNum:integer;
    TodayCharacteristic:integer;

function ReadDate(const s:string;var Date:TDate):boolean;
var l:integer; v:array[0..2] of integer;
  function SplitBy(c:char):boolean;
  var st,sp:integer; i,r:integer;
  begin
    SplitBy:=false;
    st:=1; l:=0;
    while st<=length(s) do begin
      if l>2 then exit;
      sp:=pos(c,copy(s,st,length(s)-st+1));
      if sp=0 then sp:=length(s)+2 else inc(sp,st);
      val(copy(s,st,sp-st-1),i,r);
      if r=0 then v[l]:=i else exit;
      st:=sp; inc(l);
    end;
    if (l<2) OR (l>3) then exit;
    SplitBy:=true;
  end;
begin
  ReadDate:=false;
  Date.y:=-1;
  if SplitBy('.') then begin
    Date.d:=v[0];
    Date.m:=v[1];
    if l>2 then Date.y:=v[2];
  end else if SplitBy('-') then begin
    if l>2 then begin
      Date.y:=v[0];
      Date.m:=v[1];
      Date.d:=v[2];
    end else begin
      Date.m:=v[0];
      Date.d:=v[1];
    end;
  end else if SplitBy('/') then begin
    Date.m:=v[0];
    Date.d:=v[1];
    if l>2 then Date.y:=v[2];
  end else exit;
  if (Date.y>=0) AND (Date.y<1582) then begin
    if Date.y>Today.y-2000 then inc(Date.y,1900) else inc(Date.y,2000);
  end;
  if (Date.m<1) OR (Date.d<1) OR (Date.m>12) OR (Date.d>31) then exit;
  ReadDate:=true;
end;


function DayNumber(const Date:TDate):integer;
const MonthOffsets:array[1..12]of integer=
      (0,31,59,90,120,151,181,212,243,273,304,334);
begin
  with Date do begin
    if (y<1582) OR (m<1) OR (m>12) OR (d<1) then begin
      DayNumber:=-1; exit;
    end;
    DayNumber:=d+5+
     MonthOffsets[m]+integer((m>2) AND (y AND 3=0) AND ((y MOD 100<>0) OR (y MOD 400=0)))+
     y*365+((y+3) SHR 2)-((y+99) DIV 100)+((y+399) DIV 400);
  end;
end;


function MakeDate(y,m,d:integer):TDate;
  begin Result.y:=y; Result.m:=m; Result.d:=d; end;

function GetCurrentDay:TDate;
var SysTime:TSystemTime;
begin
  GetLocalTime(SysTime);
  Result.y:=SysTime.wYear;
  Result.m:=SysTime.wMonth;
  Result.d:=SysTime.wDay;
end;


////////////////////////////////////////////////////////////////////////////////
procedure ConfigEntry(const Name,Value:string);

  procedure ParseInt(var i:integer);
  var v,r:integer;
  begin
    val(Value,v,r);
    if r=0 then i:=v else
      ParserMessageProc('illegal value "'+Value+'" for config entry "'+Name+
                        '" in the config file');
  end;

  procedure ParseWeekDays;
  var WeekDayRaw:string;
      WeekDayNo:integer;
      wd_sep_pos:integer;
  begin
    WeekDayRaw:=Value+',';
    for WeekDayNo:=1 to 7 do if WeekDayRaw<>'' then begin
      wd_sep_pos:=pos(',',WeekDayRaw);
      LC_WEEKDAYS[WeekDayNo MOD 7]:=copy(WeekDayRaw,1,wd_sep_pos-1);
      Delete(WeekDayRaw,1,wd_sep_pos);
    end;
  end;

var v,r:integer;
begin
       if Name='WarnInterval' then ParseInt(WarnInterval)
  else if Name='PrintSectionName' then ParseInt(PrintSectionName)
  else if Name='LC_LINE_DEFAULT' then LC_LINE_DEFAULT:=Value
  else if Name='LC_LINE_NO_AGE' then LC_LINE_NO_AGE:=Value
  else if Name='LC_TODAY' then LC_TODAY:=Value
  else if Name='LC_TOMORROW' then LC_TOMORROW:=Value
  else if Name='LC_THIS_WEEK' then LC_THIS_WEEK:=Value
  else if Name='LC_WEEKDAYS' then ParseWeekDays
  else if Name='LC_NEXT_WEEK' then LC_NEXT_WEEK:=Value
  else if Name='LC_SOMEDAY' then LC_SOMEDAY:=Value
  else begin
    val(Value,v,r);
    if not ParserNotifyConfigProc(Name,Value,v,r=0) then
      ParserMessageProc('illegal value "'+Value+'" for config entry "'+Name+
                        '" in the config file');
  end;
end;


////////////////////////////////////////////////////////////////////////////////
function PersonEntry(const Section,Name,Value:string):boolean;
var Date,NextBirth:TDate;
    NextBirthNum,DateDiff,Age:integer;
    DiffSpec,NameSpec,OutLine:string;
begin
  PersonEntry:=false;
  if not ReadDate(Value,Date) then exit;
  NextBirth:=Date;
  NextBirth.y:=Today.y+integer(((Date.m SHL 5) OR Date.d)<TodayCharacteristic);
  NextBirthNum:=DayNumber(NextBirth);
  DateDiff:=NextBirthNum-TodayNum;
  PersonEntry:=true;
  if DateDiff>WarnInterval then exit;
  if Date.y<0 then Age:=0 else Age:=NextBirth.y-Date.y;

  case DateDiff of
    0:DiffSpec:=LC_TODAY;
    1:DiffSpec:=LC_TOMORROW;
    2..6:DiffSpec:=Trim(StringReplace(LC_THIS_WEEK,'$weekday',LC_WEEKDAYS[NextBirthNum MOD 7],[]));
    7:DiffSpec:=LC_NEXT_WEEK;
  else
    DiffSpec:=Trim(StringReplace(LC_SOMEDAY,'$days',IntToStr(DateDiff),[]));
  end;

  if (PrintSectionName<>0)
    then NameSpec:=Name+' ['+Section+']'
    else NameSpec:=Name;

  if Age<>0 then OutLine:=LC_LINE_DEFAULT else OutLine:=LC_LINE_NO_AGE;
  OutLine:=StringReplace(OutLine,'$name',NameSpec,[]);
  OutLine:=StringReplace(OutLine,'$age',IntToStr(Age),[]);
  OutLine:=Trim(StringReplace(OutLine,'$date',DiffSpec,[]));

  SetLength(Birthdays,length(Birthdays)+1);
  with Birthdays[High(Birthdays)] do begin
    Diff:=DateDiff; Line:=OutLine;
  end;
end;


procedure SortBirthdays;
var a,b,m:integer;
    temp:TBirthday;
begin
  m:=High(Birthdays);
  for a:=0 to (m-1) do
    for b:=(a+1) to m do
      if Birthdays[a].Diff>Birthdays[b].Diff then begin
        temp:=Birthdays[a]; Birthdays[a]:=Birthdays[b]; Birthdays[b]:=temp;
      end;
end;

////////////////////////////////////////////////////////////////////////////////
procedure Parse(const FileName:string);
var t:system.text;
    Line:string; LL,SP:integer;
    Section:string;
    Identifier,Content:string;
begin
  SetLength(Birthdays,0);
  if not FileExists(FileName) then begin
    ParserMessageProc('cannot open config file ('+FileName+')');
    exit;
  end;
{$IFDEF DEBUGDAY}
  Today:=OverrideDay;
{$ELSE}
  Today:=GetCurrentDay;
{$ENDIF}
  TodayNum:=DayNumber(Today);
  TodayCharacteristic:=(Today.m SHL 5) OR Today.d;

  AssignFile(t,FileName); Reset(t);
  Section:='.';
  while not EOF(t) do begin
    readln(t,Line); Line:=Trim(Line);
    LL:=length(Line);
    if (LL>0) AND (Line[1]<>'#') AND (Line[1]<>';') then begin
      if (LL>2) AND (Line[1]='[') AND (Line[LL]=']') then
        Section:=copy(Line,2,LL-2)
      else begin
        SP:=pos('=',Line);
        if (SP>1) AND (SP<LL) then begin
          Identifier:=Trim(copy(Line,1,SP-1));
          Content:=Trim(copy(Line,SP+1,LL-SP));
          if Section[1]='.' then
            ConfigEntry(Identifier,Content)
          else
            if not PersonEntry(Section,Identifier,Content) then
              ParserMessageProc('illegal date "'+Content+
                                '" for "'+Identifier+
                                '" in section "'+Section+'" of the config file');
        end else
          ParserMessageProc('illegal line "'+Line+'" in config file');
      end;
    end;
  end;
  CloseFile(t);
  SortBirthdays;
end;

////////////////////////////////////////////////////////////////////////////////
procedure NullParserMessageProc(const msg:string); begin end;
function NullParserNotifyConfigProc(const Key,ValueS:string;ValueI:integer;IsInt:boolean):boolean;
  begin Result:=true; end;
begin
  ParserMessageProc:=NullParserMessageProc;
  ParserNotifyConfigProc:=NullParserNotifyConfigProc;
  SetLength(Birthdays,0);
end.
